from collections.abc import Sequence
from typing import Callable, Generic, Iterable, Iterator, List, Optional, Type, TypeVar

_A = TypeVar("_A")


class Seq(Generic[_A], Sequence):
    """A resuable iterable"""

    def __init__(
        self, f: Callable[[], Iterator[_A]], cached_list: Optional[List[_A]] = None
    ):
        self.f = f
        self._cached_list = cached_list

    @staticmethod
    def from_list(xs: List[_A]) -> "Seq[_A]":
        def generator() -> Iterator[_A]:
            return iter(xs)

        return Seq(generator, xs)

    def __iter__(self):
        if self._cached_list is not None:
            return iter(self._cached_list)
        else:
            return self.f()

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __getitem__(self, idx):
        return list(self)[idx]

    def __contains__(self, value: object) -> bool:
        if self._cached_list is not None:
            return value in self._cached_list
        else:
            return any(value == x for x in self.f())

    def eval(self) -> Iterable[_A]:
        if self._cached_list is None:
            self._cached_list = list(self.f())
        return self._cached_list

    def __eq__(self, other):
        if isinstance(other, Seq):
            return self.eval() == other.eval()
        elif isinstance(other, list):
            return self.eval() == other
        else:
            raise TypeError(f"Cannot compare Seq with {type(other)}")

    def __add__(self, other: Iterable[_A]) -> "Seq[_A]":
        def concat_generator() -> Iterator[_A]:
            yield from self
            yield from other

        return Seq(concat_generator)

    # ! Mutating operation is not allowed
    # def __iadd__(self, other: "Seq[_A]") -> "Seq[_A]":
    #     def concat_generator() -> Iterator[_A]:
    #         yield from self
    #         yield from other

    #     # Create a new Seq with combined generator and reassign it to self
    #     new_seq = Seq(concat_generator)
    #     self.f = new_seq.f
    #     self._cached_list = None  # Invalidate the cache
    #     return self

    def __bool__(self) -> bool:
        return any(True for _ in self)

    def __repr__(self) -> str:
        return f"Seq({list(self)})"

    def __str__(self) -> str:
        return self.__repr__()

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __copy__(self) -> "Seq[_A]":
        return Seq(self.f, self._cached_list)

    def __deepcopy__(self, memo) -> "Seq[_A]":
        return Seq(self.f, self._cached_list)

    def __reversed__(self) -> Iterator[_A]:
        if self._cached_list is not None:
            return reversed(self._cached_list)
        else:
            self._cached_list = list(self.f())
            return reversed(self._cached_list)


class _TestSeq:
    def _test_as_bool(self):
        seq = Seq.from_list([1, 2, 3])
        if seq:
            assert True
        else:
            assert False

        if not seq:
            assert False
        else:
            assert True

        empty_seq = Seq.from_list([])
        if empty_seq:
            assert False
        else:
            assert True

        if not empty_seq:
            assert True
        else:
            assert False

    def _test___add__(self):
        seq0 = Seq.from_list([])
        seq1 = Seq.from_list([1, 2, 3])
        seq2 = Seq.from_list([4, 5, 6])

        assert seq0 + seq1 == seq1
        assert seq1 + seq0 == seq1
        assert seq1 + seq2 == Seq.from_list([1, 2, 3, 4, 5, 6])



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Generic, Protocol, TypeVar

from .typeclass import Monad

_A = TypeVar("_A")
_B = TypeVar("_B")


# class _Maybe(Monad[_A], Protocol[_A]):
#     @staticmethod
#     def fmap(f: Callable[[_A], _B], x: _Maybe[_A]) -> _Maybe[_B]: ...

#     @staticmethod
#     def pure(x: _A) -> _Maybe[_A]: ...

#     @staticmethod
#     def ap(f: _Maybe[Callable[[_A], _B]], x: _Maybe[_A]) -> _Maybe[_B]: ...

#     @staticmethod
#     def bind(x: _Maybe[_A], f: Callable[[_A], _Maybe[_B]]) -> _Maybe[_B]: ...

#     def map(self, f: Callable[[_A], _B]) -> _Maybe[_B]: ...

#     def and_then(self, f: Callable[[_A], _Maybe[_B]]) -> _Maybe[_B]: ...

#     def then(self, x: _Maybe[_B]) -> _Maybe[_B]: ...

#     def unwrap(self) -> _A: ...

#     def unwrap_or(self, default: _A) -> _A: ...


@dataclass
# class Just(_Maybe[_A]):
class Just(Generic[_A]):
    value: _A

    def __repr__(self) -> str:
        return f"Just({self.value})"

    def __str__(self) -> str:
        return self.__repr__()

    def __bool__(self) -> bool:
        return True

    # def __repr__(self) -> str:
    #     return f"Just {self.value}"

    @staticmethod
    def fmap(f: Callable[[_A], _B], x: Just[_A]) -> Maybe[_B]:
        return Just(f(x.value))

    @staticmethod
    def pure(x: _A) -> Maybe[_A]:
        return Just(x)

    @staticmethod
    def ap(f: Just[Callable[[_A], _B]], x: Just[_A]) -> Maybe[_B]:
        return Just(f.value(x.value))

    @staticmethod
    def bind(x: Just[_A], f: Callable[[_A], Just[_B]]) -> Maybe[_B]:
        return f(x.value)

    def map(self, f: Callable[[_A], _B]) -> Maybe[_B]:
        return Just(f(self.value))

    def and_then(self, f: Callable[[_A], Just[_B]]) -> Maybe[_B]:
        return f(self.value)

    def then(self, x: Just[_B]) -> Maybe[_B]:
        return x

    def unwrap(self) -> _A:
        return self.value

    def unwrap_or(self, default: _A) -> _A:
        return self.value


# class Nothing(_Maybe):
@dataclass
class Nothing:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Nothing, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        pass

    def __repr__(self) -> str:
        return "Nothing"

    def __str__(self) -> str:
        return repr(self)

    def __bool__(self) -> bool:
        return False

    @staticmethod
    def fmap(f: Callable[[_A], _B], x: Nothing) -> Maybe[_B]:
        return Nothing()

    @staticmethod
    def pure(x: _A) -> Maybe[_A]:
        return Nothing()

    @staticmethod
    def ap(f: Nothing, x: Nothing) -> Maybe[Any]:
        return Nothing()

    @staticmethod
    def bind(x: Nothing, f: Callable[[_A], Nothing]) -> Maybe[_A]:
        return Nothing()

    def map(self, f: Callable[[_A], _B]) -> Maybe[_B]:
        return Nothing()

    def and_then(self, f: Callable[[_A], Maybe[_B]]) -> Nothing:
        return Nothing()

    def then(self, x: Nothing) -> Maybe[Any]:
        return Nothing()

    def unwrap(self) -> Any:
        raise ValueError("Nothing.unwrap: cannot unwrap Nothing")

    def unwrap_or(self, default: _A) -> _A:
        return default


Maybe = Just[_A] | Nothing


maybe_42 = Just(42)
maybe_42_ = Nothing()

# if __name__ == "__main__":
#     if maybe_42:
#         print(maybe_42.unwrap())
#     else:
#         print("Nothing")

#     if maybe_42_:
#         print(maybe_42_.unwrap())
#     else:
#         print("Nothing")

#     maybe_21 = maybe_42.map(lambda x: x // 2)

#     def _maybe_int(x: int) -> Maybe[int]:
#         return Just(x)

#     maybe_int = _maybe_int(42)

#     maybe_21 = maybe_int.map(lambda x: x // 2)



from pathlib import Path
import subprocess
from entoli.base.io import Io
from entoli.base.maybe import Just, Maybe, Nothing
from typing import IO, List, Tuple, Dict, Any
from dataclasses import dataclass, field


def call_command(command: str) -> Io[None]:
    def _inner() -> None:
        result = subprocess.run(command, shell=True)
        if result.returncode != 0:
            raise subprocess.CalledProcessError(result.returncode, command)

    return Io(_inner)


def call_process(program: str, args: List[str]) -> Io[None]:
    def _inner() -> None:
        result = subprocess.run([program] + args)
        if result.returncode != 0:
            raise subprocess.CalledProcessError(result.returncode, [program] + args)

    return Io(_inner)


def read_process(program: str, args: List[str], input: str = "") -> Io[str]:
    def _inner() -> str:
        result = subprocess.run(
            [program] + args,
            input=input,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        if result.returncode != 0:
            raise subprocess.CalledProcessError(
                result.returncode, [program] + args, result.stdout, result.stderr
            )
        return result.stdout

    return Io(_inner)


def read_process_with_exit_code(
    program: str, args: List[str], input: str = ""
) -> Io[Tuple[int, str, str]]:
    def _inner() -> Tuple[int, str, str]:
        result = subprocess.run(
            [program] + args,
            input=input,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        return (result.returncode, result.stdout, result.stderr)

    return Io(_inner)


# Shell interaction functions


@dataclass
class CreateProcess:
    command: str
    use_shell: bool = True
    cwd: Maybe[Path] = field(default_factory=lambda: Nothing())
    env: Maybe[Dict[str, str]] = field(default_factory=lambda: Nothing())
    std_in: Maybe[Any] = field(default_factory=lambda: Just(subprocess.PIPE))
    std_out: Maybe[Any] = field(default_factory=lambda: Just(subprocess.PIPE))
    std_err: Maybe[Any] = field(default_factory=lambda: Just(subprocess.PIPE))


# shell function
def shell(command: str) -> CreateProcess:
    return CreateProcess(command=command, use_shell=True)


# proc function
def proc(path: str, args: List[str]) -> CreateProcess:
    return CreateProcess(command=f"{path} {' '.join(args)}", use_shell=False)


# create_process function
def create_process(
    proc: CreateProcess,
) -> Io[Tuple[Maybe[IO[str]], Maybe[IO[str]], Maybe[IO[str]], subprocess.Popen]]:
    def _inner() -> (
        Tuple[Maybe[IO[str]], Maybe[IO[str]], Maybe[IO[str]], subprocess.Popen]
    ):
        process = subprocess.Popen(
            proc.command,
            shell=proc.use_shell,
            cwd=proc.cwd.unwrap_or(None),  # type: ignore
            env=proc.env.unwrap_or(None),  # type: ignore
            stdin=proc.std_in.unwrap_or(None),
            stdout=proc.std_out.unwrap_or(None),
            stderr=proc.std_err.unwrap_or(None),
            text=True,  # Ensures binary mode (bytes IO)
        )

        mb_stdin = Just(process.stdin) if process.stdin else Nothing()
        mb_stdout = Just(process.stdout) if process.stdout else Nothing()
        mb_stderr = Just(process.stderr) if process.stderr else Nothing()

        return (mb_stdin, mb_stdout, mb_stderr, process)

    return Io(_inner)


def h_put_str_ln(handle: IO[str], string: str) -> Io[None]:
    def _h_put_str_ln():
        handle.write(string + "\n")
        handle.flush()

    return Io(_h_put_str_ln)


def h_get_line(handle: IO[str]) -> Io[str]:
    def _h_get_line() -> str:
        return handle.readline().strip()

    return Io(_h_get_line)


def h_get_contents(handle: IO[str]) -> Io[str]:
    def _h_get_contents() -> str:
        return handle.read()

    return Io(_h_get_contents)


def terminate_process(process: subprocess.Popen) -> Io[None]:
    return Io(lambda: process.terminate())


def wait_for_process(process: subprocess.Popen) -> Io[int]:
    def _wait_for_process() -> int:
        return process.wait()

    return Io(_wait_for_process)


def get_process_exit_code(process: subprocess.Popen) -> Io[Maybe[int]]:
    def _inner() -> Maybe[int]:
        if code := process.poll() is None:
            return Nothing()
        else:
            return Just(code)

    return Io(_inner)



import sys
import os

# Add the src directory to the Python path
sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "src"))
)

from entoli.prelude import put_strln, get_str

greet = (
    put_strln("What is your name?")
    .then(get_str)
    .and_then(lambda name: put_strln(f"Hello, {name}!"))
)


if __name__ == "__main__":
    greet.action()



from datetime import datetime
import os
from pathlib import Path


from entoli.base.io import Io

# File system operation


def file_exists(path: Path) -> Io[bool]:
    return Io(lambda: path.exists())


def dir_exists(path: Path) -> Io[bool]:
    return Io(lambda: path.is_dir())


def list_dir(path: Path) -> Io[list[Path]]:
    return Io(lambda: list(path.iterdir()))


def read_file(path: Path) -> Io[str]:
    return Io(lambda: path.read_text())


def write_file(path: Path, content: str) -> Io[None]:
    def _inner() -> None:
        path.write_text(content)

    return Io(_inner)


def append_file(path: Path, content: str) -> Io[None]:
    def _inner() -> None:
        path.write_text(path.read_text() + content)

    return Io(_inner)


def create_dir(path: Path) -> Io[None]:
    return Io(lambda: path.mkdir())


def create_dir_if_missing(parent_as_well: bool, path: Path) -> Io[None]:
    return Io(lambda: path.mkdir(parents=parent_as_well, exist_ok=True))


def remove_file(path: Path) -> Io[None]:
    return Io(lambda: path.unlink())


def remove_dir(path: Path) -> Io[None]:
    return Io(lambda: path.rmdir())


def remove_dir_rec(path: Path) -> Io[None]:
    return Io(lambda: path.rmdir())


# todo


def get_permissions(path: Path) -> Io[os.stat_result]:
    return Io(lambda: path.stat())


def set_permissions(path: Path, mode: int) -> Io[None]:
    return Io(lambda: path.chmod(mode))


def get_modification_time(path: Path) -> Io[datetime]:
    return Io(lambda: datetime.fromtimestamp(path.stat().st_mtime))
